#include "layer_provider.hpp"

namespace horizon {
const std::map<int, Layer> &LayerProvider::get_layers() const
{
    static std::map<int, Layer> layers = {{0, {0, "Default"}}};
    return layers;
}
} // namespace horizon
